
import os
from copy import deepcopy

# Defaults for the PhantomJS Renderer
PHANTOMJS = {
    'executable': 'phantomjs',
    'args': [],
    'env': {},
    'script': os.path.join(os.path.dirname(__file__), 'render-ipc.js'),
    'timeouts': {
        'initial_page_load': 15,  # 15 Seconds, PhantomJS takes longer on the first execution after startup
        'page_load': 5,  # Max time given for PhantomJS to load the page before 'stop' and render
        'render_response': 5,  # Additional time after page load for PhantomJS to formulate and return response
        'process_startup': 10, # Max time for PhantomJS process to start before giving up
        'resource_wait_ms': 300 # Max time to wait for final resources to load, in ms
    }
}

# Defaults for the Lambda Renderer
LAMBDA = {
    'url': 'http://localhost',
    'timeouts': {
        'request_timeout': 120,  # Max time for the endpoint to finish the page render and return the results
    }
}

# Defaults for the Lifetime decorator
LIFETIME = {
    'idle_shutdown_sec': 1800,  # 30 minutes, Shutdown PhantomJS if it's been idle this long
    'max_lifetime_sec': 86400  # 24 hours, Restart PhantomJS every 24 hours
}


def merge(a, b, path=None):
    """
    Recursively merges b into a, overwriting existing a values.
    :param a:
    :type a: dict
    :param b:
    :type b: dict
    :param path:
    :return: a
    """
    if path is None: path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge(a[key], b[key], path + [str(key)])
            else:
                a[key] = deepcopy(b[key])
        else:
            a[key] = deepcopy(b[key])
    return a



import os, base64


def save_image(filename, render_response):
    """
    Saves the rendered image to a file.
    
    :param filename: The full path and filename (without extension)
    :param render_response: The response object from the renderer
    :return: True if the file was written, False otherwise
    """
    if render_response is not None and \
       u'base64' in render_response and \
       render_response[u'base64'] is not None:

        image_base64 = render_response[u'base64']
        image_bytes = base64.decodestring(image_base64)

        image_format = render_response[u'format']

        file_path = u'{file}.{format}'.format(file=filename,
                                              format=image_format.lower())

        directory, filename = os.path.split(file_path)

        if not os.path.exists(directory):
            os.makedirs(directory)

        with open(file_path, 'w') as file_stream:
            file_stream.write(image_bytes)

        return True

    return False




import abc


class Renderer(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def get_config(self):
        """ Return the configuration dictionary.
        :return: dict
        """
        raise NotImplementedError('Users must implement get_config() to use this base class')

    @abc.abstractmethod
    def render(self, url, html=None, img_format='PNG', width=1280, height=1024, page_load_timeout=None, user_agent=None,
               headers=None, cookies=None, html_encoding=u'utf-8'):
        raise NotImplementedError('Users must implement render() to use this base class')

    @abc.abstractmethod
    def shutdown(self, timeout=None):
        """ Shutdown and cleanup the renderer.
        :param timeout: Time in seconds to wait for shutdown.
        :return:
        """
        raise NotImplementedError('Users must implement shutdown() to use this base class')


class RenderError(Exception):
    """Exception raised for errors during rendering.

    Attributes:
        msg  -- explanation of the error
    """
    def __init__(self, msg, *args, **kwargs):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)


SCHEMA = {
    "1.0": {
        "render": {
            "properties": {
                "url": {
                    "type": "string",
                    "minLength": 3
                },
                "html": {
                    # default is None
                    "type": "string",
                    "minLength": 1,
                    "binaryEncoding": "base64",
                    "contentMediaType": "text/html"
                },
                "img_format": {
                    # default here is PNG
                    "type": "string",
                    "enum": [
                        "PDF",
                        "PNG",
                        "JPEG",
                        "BMP",
                        "PPM",
                    ]
                },
                "width": {
                    # default 1280
                    "type": "number",
                    "minimum": 1
                },
                "height": {
                    # default 1024
                    "type": "number",
                    "minimum": 1
                },
                "page_load_timeout": {
                    # default None
                    "type": "number",
                    "minimum": 1
                },
                "user_agent": {
                    # default None
                    "type": "string",
                    "minLength": 1
                },
                "headers": {
                    # default None
                    "type": "object"
                },
                "cookies": {
                    # default None
                    "type": "object"
                },
                "html_encoding": {
                    # default utf-8
                    "type": "string",
                    "minLength": 1
                },
            },
            "required": [
                "url"
            ],
            "additionalProperties": False
        }
    }
}


from renderer import Renderer
import requests
from requests.exceptions import RequestException, \
                                ConnectionError, \
                                Timeout, \
                                TooManyRedirects
import base64
import copy
import logging
from settings import LAMBDA, merge
import traceback
import time

class LambdaRenderer(Renderer):
    """Offloads the rendering process to a PhantomJSRenderer
    running inside of AWS Lambda
    """

    def __init__(self, config, logger=None):

        self.config = copy.deepcopy(LAMBDA)
        self.config = merge(self.config, config)

        if logger is not None:
            self._logger = logger
        else:
            self._logger = logging.getLogger(u'LambdaRenderer')

    def get_config(self):
        """ Return the configuration dictionary.
        :return: dict
        """
        return self.config

    def is_base64(self, the_string):
        try:
            return base64.b64encode(base64.b64decode(the_string)) == the_string
        except Exception:
            return False

    def _prep_json(self, url, html, img_format, width, height, page_load_timeout,
                   user_agent, headers, cookies, html_encoding):
        """Preps the json value to be passed to the request"""
        json_dict = {
        # non-None args
            'url': url,
            'img_format': img_format,
            'width': width,
            'height': height,
            'html_encoding': html_encoding,
        }

        # args that can be None
        if html is not None:
            # accept both text and base64 encoded text
            if self.is_base64(html):
                json_dict['html'] = html
            else:
                json_dict['html'] = base64.b64encode(html)

        if page_load_timeout is not None:
            json_dict['page_load_timeout'] = page_load_timeout

        if user_agent is not None:
            json_dict['user_agent'] = user_agent

        if headers is not None:
            json_dict['headers'] = headers

        if cookies is not None:
            json_dict['cookies'] = cookies

        return json_dict

    def _prep_headers(self):
        """Preps the headers to send to lambda"""
        h_dict = {
            'content-type':'application/json'
        }

        if 'api_key' in self.config and \
                self.config['api_key'] is not None:
            h_dict['x-api-key'] = self.config['api_key']

        return h_dict

    def _prep_timeout(self):
        """Preps the request timeout to lambda"""
        return self.config['timeouts']['request_timeout']

    def render(self, url, html=None, img_format='PNG', width=1280, height=1024,
               page_load_timeout=None, user_agent=None,
               headers=None, cookies=None, html_encoding=u'utf-8'):
        """
        Render a URL target or HTML to an image file.
        :param url:
        :param html:
        :param img_format:
        :param width:
        :param height:
        :param page_load_timeout:
        :param user_agent:
        :param headers:
        :param cookies:
        :param html_encoding:
        :return dict:
        """
        # prep request
        json_dict = self._prep_json(url=url,
                                    html=html,
                                    img_format=img_format,
                                    width=width,
                                    height=height,
                                    page_load_timeout=page_load_timeout,
                                    user_agent=user_agent,
                                    headers=headers,
                                    cookies=cookies,
                                    html_encoding=html_encoding)
        request_headers = self._prep_headers()
        timeout = self._prep_timeout()

        # send it
        try:
            self._logger.info("Sending lambda request {} {} {} {}".format(url, json_dict, request_headers, timeout))
            start_time = time.time()
            result = requests.post(url=self.config['url'],
                                   json=json_dict,
                                   headers=request_headers,
                                   allow_redirects=True,
                                   timeout=timeout)
            self._logger.info("Request took {}s".format(time.time() - start_time))

            # valid response should be json
            json_result = result.json()
            json_copy = copy.deepcopy(json_result)
            if 'base64' in json_copy and json_copy['base64'] is not None:
                json_copy['base64'] = '<omitted>'
            self._logger.debug("Received data from lambda {}".format(json_copy))
        except (ValueError, ConnectionError, Timeout,
                TooManyRedirects, RequestException) as e:
            self._logger.error("Exception while calling lambda {}".format(traceback.format_exc()))
            return {
                u'url': url,
                u'status': u'fail',
                u'load_time': None,
                u'paint_time': None,
                u'base64': None,
                u'format': img_format,
                u'error': traceback.format_exc(),
            }

        # handle error within lambda function
        if result.status_code != 200:
            self._logger.warn("Received unexpected {} status code from lambda".format(result.status_code))
            response = {
                u'url': url,
                u'status': u'fail',
                u'load_time': None,
                u'paint_time': None,
                u'base64': None,
                u'format': img_format,
                u'error': None
            }

            if 'message' in json_result:
                response['error'] = json_result['ex'] if 'ex' in json_result else json_result['message']
            return response

        return json_result

    def shutdown(self, timeout=None):
        pass



#!/usr/bin/env python

import eventlet
from eventlet.green import subprocess
from eventlet.green import threading
from eventlet.timeout import Timeout
from eventlet.queue import Queue, Empty

import copy
import json
import os
import base64
import traceback
import time
import renderer
import logging

from signal import *
from settings import PHANTOMJS, merge


class PhantomJSRenderer(renderer.Renderer):
    """
    Render a web page to an image, using either a URL or raw HTML.

    Requires PhantomJS to be installed on the host: http://phantomjs.org/
    """

    def __init__(self, config, logger=None, register_shutdown=False):

        self.config = copy.deepcopy(PHANTOMJS)
        self.config = merge(self.config, config)

        self._proc = None
        self._stderr_reader = None
        self._comms_lock = threading.RLock()
        self._shutdown_lock = threading.RLock()

        if not self._which(self.config[u'executable']):
            raise renderer.RenderError(''.join([u"Can't locate PhantomJS executable: ", self.config[u'executable']]))

        if not os.path.isfile(self.config[u'script']):
            raise renderer.RenderError(''.join([u"Can't locate script: ", self.config[u'script']]))

        if logger is not None:
            self._logger = logger
        else:
            self._logger = logging.getLogger(u'PhantomJSRenderer')

        if register_shutdown and isinstance(threading.current_thread(), threading._MainThread):
            for sig in (SIGABRT, SIGINT, SIGTERM):
                signal(sig, self._on_signal)

    def get_config(self):
        return self.config

    def render(self, url, html=None, img_format=u'PNG', width=1280, height=1024, page_load_timeout=None, user_agent=None,
               headers=None, cookies=None, html_encoding=u'utf-8'):
        """
        Render a URL target or HTML to an image file.
        :param url:
        :param html:
        :param img_format:
        :param width:
        :param height:
        :param page_load_timeout:
        :param user_agent:
        :param headers:
        :param cookies:
        :param html_encoding:
        :return:
        """

        request = {u'url': url, u'width': width, u'height': height, u'format': img_format}

        if html is not None:
            if isinstance(html, unicode):
                html = html.encode(html_encoding, errors='replace')

            b64 = base64.b64encode(html)
            request[u'html64'] = b64

        if user_agent is not None:
            request[u'userAgent'] = user_agent

        if headers is not None:
            request[u'headers'] = headers

        if cookies is not None:
            request[u'cookies'] = cookies

        with self._comms_lock:
            try:
                first_render = False

                if not hasattr(self, '_proc') or self._proc is None:
                    startup_timeout = self.config[u'timeouts'][u'process_startup']

                    command = self._construct_command()

                    self._logger.info(u"Starting the PhantomJS process: " + u" ".join(command))

                    with Timeout(startup_timeout):
                        self._proc = subprocess.Popen(command,
                                                      bufsize=4096,
                                                      shell=False,
                                                      stdin=subprocess.PIPE,
                                                      stdout=subprocess.PIPE,
                                                      stderr=subprocess.PIPE,
                                                      env=self.config[u'env'])

                        self._stderr_reader = PipeReader(self._proc.stderr)

                    first_render = True

                render_timeout = self.config[u'timeouts'][u'render_response']

                if page_load_timeout is None:
                    if first_render:
                        page_load_timeout = self.config[u'timeouts'][u'initial_page_load']
                    else:
                        page_load_timeout = self.config[u'timeouts'][u'page_load']

                request[u'timeout'] = page_load_timeout * 1000  # Convert seconds to ms
                request[u'resourceWait'] = self.config[u'timeouts'][u'resource_wait_ms']

                request_string = json.dumps(request)

                try:
                    with Timeout(page_load_timeout + render_timeout):

                        self._logger.debug(u'Sending request: ' + request_string)
                        self._proc.stdin.write(request_string + '\n')
                        self._proc.stdin.flush()

                        response_string = self._proc.stdout.readline()

                except Timeout:
                    response_string = None
                    self._logger.warn(u'Received no response, terminating PhantomJS.')
                    self.shutdown()

                err_messages = self._check_stderr()

                if err_messages is not None and self._logger.isEnabledFor(logging.DEBUG):
                    self._logger.debug(err_messages)

                response = {u'url': url,
                            u'status': None,
                            u'load_time': None,
                            u'paint_time': None,
                            u'base64': None,
                            u'format': img_format,
                            u'error': None}

                if response_string is None:
                    response[u'status'] = u'fail'
                    response[u'error'] = u'Render request has timed out.'
                else:
                    try:
                        phantom_response = json.loads(response_string)

                        if self._logger.isEnabledFor(logging.DEBUG):
                            msg = {}
                            msg.update(phantom_response)
                            msg[u'base64'] = u'<omitted>'
                            self._logger.debug(u'Received response: ' + json.dumps(msg))

                        if u'status' in phantom_response:
                            response[u'status'] = phantom_response[u'status']
                        else:
                            response[u'status'] = u'fail'

                        if u'loadTime' in phantom_response:
                            response[u'load_time'] = phantom_response[u'loadTime']

                        if u'paintTime' in phantom_response:
                            response[u'paint_time'] = phantom_response[u'paintTime']

                        if u'base64' in phantom_response:
                            response[u'base64'] = phantom_response[u'base64']

                        if u'error' in phantom_response:
                            response[u'error'] = json.dumps(phantom_response[u'error'])
                        elif err_messages is not None:
                            response[u'error'] = err_messages

                    except (ValueError, KeyError) as e:
                        self._logger.debug(u'Error parsing response: {}\nTerminating PhantomJS.\n{}'.format(response_string, traceback.format_exc()))

                        self.shutdown()

                        response[u'status'] = u'fail'
                        response[u'error'] = ''.join([str(e), u'\nPhantomJS response: ', response_string])

                return response

            except (Timeout, Exception):
                self._logger.error(u'Unexpected error, terminating PhantomJS.\n' + traceback.format_exc())
                self.shutdown()
                raise

    def shutdown(self, timeout=None):
        """

        :return:
        """
        with self._shutdown_lock:
            if hasattr(self, '_proc') and self._proc is not None:
                try:
                    self._proc.kill()
                    self._proc.wait()
                finally:
                    del self._proc

                self._logger.info(u'PhantomJS terminated.')

        if self._stderr_reader is not None:
            self._stderr_reader.shutdown()

    def _check_stderr(self):
        """Collect any input from the stderr pipe and return it."""

        err = self._stderr_reader.get()
        err_messages = []

        while err is not None:
            err_messages.append(err.decode('UTF-8', errors='replace'))
            err = self._stderr_reader.get()

        if len(err_messages) > 0:
            return u'\n'.join(err_messages)
        else:
            return None

    def _on_signal(self, sig, frame):
        """"""
        self.shutdown()
        os._exit(0)

    def _construct_command(self):
        """Build the command array for executing the PhantomJS process."""

        executable = self._which(self.config[u'executable'])

        return [executable] + self.config[u'args'] + [self.config[u'script']]

    @staticmethod
    def _which(program):
        """Locate the executable on the path if required."""

        def is_exe(fpath):
            return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

        fpath, fname = os.path.split(program)
        if fpath:
            if is_exe(program):
                return program
        else:
            for path in os.environ["PATH"].split(os.pathsep):
                path = path.strip('"')
                exe_file = os.path.join(path, program)
                if is_exe(exe_file):
                    return exe_file

        return None


class PipeReader:

    def __init__(self, pipe):

        self._pipe = pipe
        self._queue = Queue()

        _queue = self._queue

        def _enqueue_output():
            for line in iter(pipe.readline, b''):
                _queue.put(line)

            pipe.close()

        self._green_thread = eventlet.spawn(_enqueue_output)

    def get(self):
        try:
            return self._queue.get_nowait()
        except Empty:
            pass
        return None

    def shutdown(self):

        if self._green_thread is not None:
            try:
                self._green_thread.kill()
            except:
                pass




import eventlet
from eventlet.green import threading

import copy
import logging
import time

from renderer import Renderer
from settings import LIFETIME


class Lifetime(Renderer):
    """
    Wraps a Renderer instance and limits the maximum time a single render
    instance can live, and also shuts down a process if it has been unused for
    a specified time period.

    PhantomJS can be configured to use an in-memory browser cache which can
    grow without bounds. The Lifetime decorator is useful for controlling this
    growth by limiting the amount of time a single Renderer can live, and
    releasing its accumulated resources.

    For long or continual running of the rendering process, it is highly
    recommended to wrap the Renderer with the Lifetime decorator to prevent
    eventual OOM.
    """

    _delegate = None
    _last_render_time = None
    _start_time = None

    _running = False
    _thread = None
    _condition = threading.Condition(threading.RLock())
    _lock = threading.RLock()

    def __init__(self, renderer):

        self._delegate = renderer
        self.config = copy.deepcopy(LIFETIME)
        self.config.update(renderer.get_config())

        self._logger = logging.getLogger(u'LifetimeDecorator')

    def get_config(self):
        return self.config

    def render(self, url, html=None, img_format='PNG', width=1280, height=1024, page_load_timeout=None, user_agent=None,
               headers=None, cookies=None, html_encoding=u'utf-8'):

        with self._lock:
            self._last_render_time = time.time()

            if self._start_time is None:
                self._start_time = self._last_render_time

            if not self._running:
                self._startup()

            return self._delegate.render(url, html, img_format, width, height, page_load_timeout, user_agent, headers, cookies, html_encoding)

    def _startup(self):

        self._running = True
        self._thread = eventlet.spawn(self._lifetime_monitor)

    def shutdown(self, timeout=None):

        self._running = False

        if self._condition is not None:
            # Wake the monitor thread
            with self._condition:
                self._condition.notify()

        if self._thread is not None:
            try:
                self._thread.kill()
            except Exception:
                pass

        self._delegate.shutdown(timeout)

    def _lifetime_monitor(self):

        while self._running:

            now = time.time()
            sleep_delta = None

            with self._lock:
                if self._last_render_time is not None:

                    idle_target = self.config['idle_shutdown_sec'] + self._last_render_time

                    if now >= idle_target:
                        self._logger.info(u"Shutting down idle renderer.")

                        self._last_render_time = None
                        self._start_time = None
                        self._delegate.shutdown()
                        self._running = False
                        break
                    else:
                        sleep_delta = idle_target - now

                if self._start_time is not None:

                    expired_target = self.config['max_lifetime_sec'] + self._start_time

                    if now >= expired_target:
                        self._logger.info(u"Shutting down renderer which reached max lifetime.")

                        self._last_render_time = None
                        self._start_time = None
                        self._delegate.shutdown()
                        self._running = False
                        break
                    else:
                        if sleep_delta is not None:
                            sleep_delta = min(sleep_delta, expired_target - now)
                        else:
                            sleep_delta = expired_target - now

            if sleep_delta is not None:
                with self._condition:
                    if self._running:
                        self._condition.wait(sleep_delta)


#!/usr/bin/env python
# encoding: utf-8

from setuptools import setup

import re, os


def get_version():
    with open('phantom_snap/__version__.py') as version_file:
        return re.search(r"""__version__\s+=\s+(['"])(?P<version>.+?)\1""",
            version_file.read()).group('version')

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()

setup(name='phantom-snap',
    description='Render HTML to an image using PhantomJS with this library \
                  designed to scale for high volume continuous operation.',
    long_description=README,
    author='Andrew Carter',
    author_email='andrew.carter@istresearch.com',
    url='https://github.com/istresearch/phantom-snap',
    keywords=['web', 'html', 'render', 'screenshot', 'phantomjs'],
    license='MIT',
    version=get_version(),
    packages=[
        'phantom_snap'
    ],
    package_data={
        'phantom_snap': ['*.js']
    },
    setup_requires=[
        'nose>=1.3.7'
    ],
    install_requires=[
        'eventlet>=0.23.0',
        'requests>=2.19.1'
    ],
    tests_require=[
        'nose',
        'coverage>=4.0.3',
        'mock==2.0.0',
        'jsonschema>=2.6.0',
    ],
    test_suite='nose.collector',
    classifiers=[
        "Intended Audience :: Developers",
        "Topic :: System",
        "Topic :: System :: Distributed Computing",
        "Topic :: Utilities"
    ]
    )
